import torch
from torch.nn.functional import normalize

from .base import LossBase


class ContrastiveLoss(torch.nn.Module):
    def __init__(
        self,
        negative_samples=2048,
        temperature=0.5,
        loss_mode="infonce",
        metric="euclidean",
        eps=1.0,
        noise_in_estimator=1.0,
        clamp_lo=float("-inf"),
        clamp_hi=float("inf"),
        seed=0,
        loss_aggregation="mean",
    ):
        super(ContrastiveLoss, self).__init__()
        self.negative_samples = negative_samples
        self.temperature = temperature
        self.loss_mode = loss_mode
        self.metric = metric
        self.noise_in_estimator = noise_in_estimator
        self.eps = eps
        self.clamp_lo = clamp_lo
        self.clamp_hi = clamp_hi
        self.clamp = self.clamp_lo, self.clamp_hi
        self.seed = seed
        torch.manual_seed(self.seed)
        self.neigh_inds = None
        self.loss_aggregation = loss_aggregation

        if self.loss_mode == "nce":
            self.log_Z = torch.tensor(0.0)
            self.log_Z = torch.nn.Parameter(self.log_Z, requires_grad=True)

    def forward(
        self,
        features,
        backbone_features=None,
        labels=None,
        force_resample=False,
    ):
        """Compute loss for model. SimCLR unsupervised loss:
        https://arxiv.org/pdf/2002.05709.pdf

        Args:
            features: hidden vector of shape [2 * bsz, n_views, ...].
            force_resample: Whether the negative samples should be forcefully resampled.
        Returns:
            A loss scalar.
        """

        batch_size = features.shape[0] // 2
        b = batch_size

        # We can at most sample this many samples from the batch.
        # `b` can be lower than `self.negative_samples` in the last batch.
        negative_samples = min(self.negative_samples, 2 * b - 1)

        if force_resample or self.neigh_inds is None:
            neigh_inds = make_neighbor_indices(
                batch_size, negative_samples, device=features.device
            )
            self.neigh_inds = neigh_inds
        # untested logic to accomodate for last batch
        elif self.neigh_inds.shape[0] != batch_size:
            neigh_inds = make_neighbor_indices(
                batch_size, negative_samples, device=features.device
            )
            # don't save this one
        else:
            neigh_inds = self.neigh_inds
        neighbors = features[neigh_inds, :]

        # `neigh_mask` indicates which samples feel attractive force
        # and which ones repel each other
        neigh_mask = torch.ones_like(neigh_inds, dtype=torch.bool)
        neigh_mask[:, 0] = False

        origs = features[:b]

        # compute probits
        if self.metric == "euclidean":
            dists = ((origs[:, None] - neighbors) ** 2).sum(axis=2)
            # Cauchy affinities
            probits = torch.div(1, self.eps + dists)
        elif self.metric == "cosine":
            o = normalize(origs.unsqueeze(1), dim=2)
            n = normalize(neighbors.transpose(1, 2), dim=1)
            logits = torch.bmm(o, n).squeeze() / self.temperature
            # logits_max, _ = logits.max(dim=1, keepdim=True)
            # logits -= logits_max.detach()
            # logits -= logits.max().detach()
            probits = torch.exp(logits)
        else:
            raise ValueError(f"Unknown metric “{self.metric}”")

        # compute loss
        if self.loss_mode == "nce":
            self.log_Z.to(features.device)

            # for proper nce it should be negative_samples *
            # p_noise. But for uniform noise distribution we would
            # need the size of the dataset here. Also, we do not use a
            # uniform noise distribution as we sample negative samples
            # from the batch.

            if self.metric == "euclidean":
                # estimator is (cauchy / Z) / ( cauchy / Z +
                # neg_samples)). For numerical stability rewrite to 1
                # / ( 1 + (d**2 + eps) * Z * m)
                estimator = 1 / (
                    1
                    + (dists + self.eps)
                    * torch.exp(self.log_Z)
                    * negative_samples
                )
            else:
                probits = probits / torch.exp(self.log_Z)
                estimator = probits / (probits + negative_samples)

            loss = -(~neigh_mask * torch.log(estimator.clamp(*self.clamp))) - (
                neigh_mask * torch.log((1 - estimator).clamp(*self.clamp))
            )
        elif self.loss_mode == "neg_sample":
            if self.metric == "euclidean":
                # estimator rewritten for numerical stability as for nce
                estimator = 1 / (
                    1 + self.noise_in_estimator * (dists + self.eps)
                )
            else:
                estimator = probits / (probits + self.noise_in_estimator)

            loss = -(~neigh_mask * torch.log(estimator.clamp(*self.clamp))) - (
                neigh_mask * torch.log((1 - estimator).clamp(*self.clamp))
            )

        elif self.loss_mode == "umap":
            # cross entropy parametric umap loss
            loss = -(~neigh_mask * torch.log(probits.clamp(*self.clamp))) - (
                neigh_mask * torch.log((1 - probits).clamp(*self.clamp))
            )
        elif self.loss_mode == "infonce":
            # loss from e.g. sohn et al 2016, includes pos similarity
            # in denominator
            loss = -(
                (torch.log(probits.clamp(*self.clamp)[~neigh_mask]))
                - torch.log(probits.clamp(*self.clamp).sum(axis=1))
            )
        elif self.loss_mode == "infonce_alt":
            # loss simclr
            loss = -(
                (torch.log(probits.clamp(*self.clamp)[~neigh_mask]))
                - torch.log(
                    (neigh_mask * probits.clamp(*self.clamp)).sum(axis=1)
                )
            )
        else:
            raise ValueError(f"Unknown loss_mode “{self.loss_mode}”")

        # aggregate loss over batch
        if self.loss_aggregation == "sum":
            loss = loss.sum()
        else:
            loss = loss.mean()

        return loss


def make_neighbor_indices(batch_size, negative_samples, device=None):
    """
    Selects neighbor indices
    :param batch_size: int Batch size
    :param negative_samples: int Number of negative samples
    :param device: torch.device Device of the model
    :return: torch.tensor Neighbor indices
    :rtype:
    """
    b = batch_size

    if negative_samples < 2 * b - 1:
        # uniform probability for all points in the minibatch,
        # we sample points for repulsion randomly
        neg_inds = torch.randint(
            0, 2 * b - 1, (b, negative_samples), device=device
        )
        neg_inds += (torch.arange(1, b + 1, device=device) - 2 * b)[:, None]
    else:
        # full batch repulsion
        all_inds1 = torch.repeat_interleave(
            torch.arange(b, device=device)[None, :], b, dim=0
        )
        not_self = ~torch.eye(b, dtype=bool, device=device)
        neg_inds1 = all_inds1[not_self].reshape(b, b - 1)

        all_inds2 = torch.repeat_interleave(
            torch.arange(b, 2 * b, device=device)[None, :], b, dim=0
        )
        neg_inds2 = all_inds2[not_self].reshape(b, b - 1)
        neg_inds = torch.hstack((neg_inds1, neg_inds2))

    # now add transformed explicitly
    neigh_inds = torch.hstack(
        (torch.arange(b, 2 * b, device=device)[:, None], neg_inds)
    )

    return neigh_inds


class SlowContrastiveLoss(LossBase):
    def get_deps(self):
        supdeps = super().get_deps()
        # the default return value for kwargs.get needs to match the
        # "loss_mode" kwarg in the ContrastiveLoss class
        if self.kwargs.get("loss_mode", "infonce") == "nce":
            deps = [self.indir / "model.pt"]
        else:
            deps = []
        return supdeps + deps

    def load(self):
        if self.kwargs.get("loss_mode", "infonce") == "nce":
            self.model_sd = torch.load(self.indir / "model.pt")

    def compute(self):
        self.seed = self.random_state.integers(2**63 - 1)
        self.criterion = ContrastiveLoss(
            metric=self.metric, seed=self.seed, **self.kwargs
        )

        # Add the normalization constant to the optimizer
        if self.kwargs.get("loss_mode", "infonce") == "nce":
            self.opt = self.model_sd["opt"]
            self.opt.add_param_group(dict(params=self.criterion.log_Z))

    def save(self):
        super().save()

        if self.kwargs.get("loss_mode", "infonce") == "nce":
            self.model_sd["opt"] = self.opt
            self.model_sd["opt_sd"] = self.opt.state_dict()
            self.save_lambda_alt(
                self.outdir / "model.pt", self.model_sd, torch.save
            )



import sys
import time

import numpy as np
import torch


def train(
    train_loader,
    model,
    log_Z,
    criterion,
    optimizer,
    epoch,
    clip_grad=True,
    print_freq=None,
    force_resample=None,
):
    """
    one epoch training
    :param train_loader: DataLoader Returns batches of similar tuples
    :param model: torch.nn.Module Embedding layer (non-parametric) or neural network (parametric)
    :param log_Z: torch.tensor Float containing the logarithm of the learnable Z
    :param criterion: torch.nn.Module Computes the loss
    :param optimizer:  torch.optim.Optimizer
    :param epoch: int Current training epoch
    :param clip_grad: bool If True, clips gradients to 4
    :param print_freq: int or None Frequency for printing if not None
    :param force_resample: bool or None If True, forces resampling of negative sample indices for every batch. If None, once every epoch.
    :return: torch.tensor Losses for all batches of the epoch
    """
    model.train()
    losses = []
    for idx, (item, neigh) in enumerate(train_loader):
        print_now = print_freq is not None and (idx + 1) % print_freq == 0
        start = time.time()

        images = torch.cat([item, neigh], dim=0)

        images = images.to(next(model.parameters()).device)

        # compute loss
        features = model(images)
        if print_now:
            features.retain_grad()  # to print model agnostic grad statistics
        force_resample = force_resample if force_resample is not None else idx == 0
        loss = criterion(features, log_Z, force_resample=force_resample)

        # update metric
        losses.append(loss.item())

        # Update parameters
        optimizer.zero_grad()
        loss.backward()
        if clip_grad:
            torch.nn.utils.clip_grad_value_(model.parameters(), 4)
            if log_Z is not None:
                torch.nn.utils.clip_grad_value_(log_Z, 4)
        optimizer.step()

        # print info
        if print_now:
            print(
                f"Train: E{epoch}, {idx}/{len(train_loader)}\t"
                # print grad on features to be model agnostic
                f"grad magn {features.grad.abs().sum():.3f}, "
                f"loss {sum(losses) / len(losses):.3f}, "
                f"time/iteration {time.time() - start:.3f}",
                file=sys.stderr,
            )
            if torch.isnan(features).any() or torch.isnan(loss).any():
                print(
                    f"NaN error! feat% {torch.isnan(features).sum() / (features.shape[0] * features.shape[1]):.3f}, "
                    f"loss% {torch.isnan(loss).sum():.3f}",
                    file=sys.stderr,
                )
                exit(3)

    return losses


class ContrastiveEmbedding(object):
    """
    Class for computing contrastive embeddings from similarity information.
    """

    def __init__(
        self,
        model: torch.nn.Module,
        batch_size=2**15,
        negative_samples=5,
        n_epochs=200,
        device="auto",
        learning_rate=0.001,
        lr_min_factor=0.0,
        momentum=0.0,
        temperature=0.5,
        noise_in_estimator=None,
        neg_spec=None,
        ince_spec=None,
        Z_bar=None,
        s=None,
        eps=1.0,
        clamp_high="auto",
        clamp_low="auto",
        Z=1.0,
        loss_mode="infonce",
        metric="euclidean",
        optimizer="sgd",
        weight_decay=0,
        anneal_lr="none",
        lr_decay_rate=0.1,
        lr_decay_epochs=None,  # unused for now
        clip_grad=True,
        save_freq=25,
        callback=None,
        print_freq_epoch="auto",
        print_freq_iteration=None,
        seed=0,
        loss_aggregation="sum",
        force_resample=None,
        warmup_epochs=0,
        warmup_lr=0,
        early_exaggeration=True
    ):
        """
        :param model: torch.nn.Module Embedding model (embedding layer for non-parametric, neural network for parametric)
        :param batch_size: int Batch size
        :param negative_samples: int Number of negative samples per positive sample
        :param n_epochs: int Number of optimization epochs
        :param device: torch.device or "auto" Device of optimization. If auto, cuda is used if available.
        :param learning_rate: float Learning rate
        :param lr_min_factor: float Minimal value to which learning rate is annealed
        :param momentum: float Momentum of SGD
        :param temperature: float Temperature used in Cosine similarity
        :param noise_in_estimator: float Value used in negative sampling's fraction q / (q+ noise_in_estimator), redundant with Z_bar. Deprecated, use neg_spec instead.
        :param neg_spec: float Value used in negative sampling's fraction q / (q+ neg_spec). Controls the point on the negative sampling spectrum. Redundant with Z_bar, which has priority.
        :param ince_spec: float The repulsive term in the InfoNCE loss is multiplied by the inverse of ince_spec. Controls the InfoNCE spectrum. Redundant with s, which has priority.
        :param Z_bar: float Fixed normalization constant in negative sampling. Controls the negative sampling spectrum. Redundant with s, which has priority.
        :param s: float Slider parameter for the spectrum of neighbor embeddings. If s=1, the embedding will be similar to a UMAP embedding. If s=0, the embedding will be similar to a t-SNE embedding. Logarithmic inter-/ extrapolation. Works for the loss modes "infonce", "infonce_alt" and "neg".
        :param eps: float Iterpolates between UMAP's implicit similarity (eps = 0) and the Cauchy kernels (eps = 1.0)
        :param clamp_high: float Upper value at which arguments to logarithms are clamped. Default "auto" chooses values based on the metric. For metric="euclidean" it is 1.0, for metric="cosine" it is inf.
        :param clamp_low: float Lower value at which arguments to logarithms are clamped. Default "auto" chooses values based on the metric. For metric="euclidean" it is 1e-4, for metric="cosine" it is -inf.
        :param Z: float Initial value for the learned normalization parameter of NCE
        :param loss_mode: str Specifies which loss to use. Must be one of "umap", "neg", "nce", "infonce", "infonce_alt". "neg_sample" is deprecated and defaults to "neg"
        :param metric: str Specifies which metric to use for computing distances. Must be "cosine" or "euclidean".
        :param optimizer: str Specifies which optimizer to use. Must be "sgd" or "adam"
        :param weight_decay: float Value of weight decay.
        :param anneal_lr: bool If True, the learning rate is annealed
        :param lr_decay_rate: float Parameter for speed of learing rate decay
        :param lr_decay_epochs: int Number of epochs over which learning rate is decayed
        :param clip_grad: bool If True, gradients are clipped
        :param save_freq: int Frequency in epochs of calling callback.
        :param callback: callable Callback to call before first and every save_freq epochs.
        :param print_freq_epoch: int Epoch progress is printed every print_freq_epoch epoch
        :param print_freq_iteration: int Losses are printed every print_freq_iteration batch per epoch
        :param seed: int Random seed
        :param loss_aggregation: str Specifies how to aggregate loss over a batch. Must be "sum" or "mean".
        :param force_resample: bool or None If True, negative sample indices are resampled every batch. If None, they are resampled every epoch.
        :param warmup_epochs: int Number of epochs for linearly warming up the learning rate
        :param warmup_lr: float Starting learning rate to warm up from.
        :param early_exaggeration: bool Whether to use the first third of the optimization in the s=1 regime. Only affects loss modes "infonce" and "neg".
        """
        self.model: torch.nn.Module = model
        self.batch_size: int = batch_size
        self.negative_samples: int = negative_samples
        self.n_epochs: int = n_epochs


        if device == "auto":
            cuda_available = torch.cuda.is_available()
            if cuda_available:
                device = "cuda"
            else:
                device = "cpu"
        self.device = device
        self.learning_rate = learning_rate
        self.momentum = momentum
        self.temperature = temperature
        self.loss_mode: str = loss_mode
        self.metric: str = metric
        self.optimizer = optimizer
        self.weight_decay = weight_decay
        if isinstance(anneal_lr, bool):
            anneal_lr = "linear" if anneal_lr else "none"
        self.anneal_lr: str = anneal_lr
        self.lr_min_factor: float = lr_min_factor
        self.lr_decay_rate = lr_decay_rate
        self.lr_decay_epochs = lr_decay_epochs
        self.warmup_lr = warmup_lr
        self.warmup_epochs = warmup_epochs
        self.clip_grad: bool = clip_grad
        self.save_freq: int = save_freq
        self.callback = callback

        if print_freq_epoch == "auto":
            self.print_freq_epoch = self.n_epochs // 5
        else:
            self.print_freq_epoch = print_freq_epoch
        self.print_freq_iteration = print_freq_iteration
        self.eps = eps
        self.seed = seed
        self.loss_aggregation = loss_aggregation
        self.force_resample = force_resample
        self.warmup_epochs = warmup_epochs
        self.warmup_lr = warmup_lr

        # alias for loss mode "neg" to ensure backwards compatibility
        # still support "neg_sample" since the loss mode is put into the file names, which are featured in the notebooks
        if self.loss_mode == "neg_sample":
            self.loss_mode = "neg"

        if self.loss_mode == "nce":
            self.log_Z = torch.tensor(np.log(Z), device=self.device)
            self.log_Z = torch.nn.Parameter(self.log_Z, requires_grad=True)
            early_exaggeration = False
        else:
            self.log_Z = None

        if self.loss_mode == "neg":
            n_specified_params = (noise_in_estimator is not None) + (Z_bar is not None) + (s is not None) + (neg_spec is not None)
            if n_specified_params == 0:
                s = 1.0  # default for neg is umap

            if n_specified_params > 1:
                print(
                    "Warning: More than one of 'noise_in_estimator', 'neg_spec', 'Z_bar' and "
                    "'s' were specified. 's' will supersede 'Z_bar', which supersedes 'neg_spec', which supersedes 'noise_in_estimator'."
                )

        if self.loss_mode == "infonce" or self.loss_mode == "infonce_alt":
            n_specified_params = (noise_in_estimator is not None) + (ince_spec is not None) + (s is not None)
            if n_specified_params == 0:
                s = 0.0  # default for infonce is tsne

            if n_specified_params > 1:
                print(
                    "Warning: More than one of 'noise_in_estimator', 'ince_spec', and "
                    "'s' were specified. 's' will supersede 'ince_spec', which supersedes 'noise_in_estimator'."
                )

        # set up clamping values depending on loss mode
        if clamp_low == "auto":
            if self.metric == "euclidean":
                self.clamp_low = 1e-4
            elif self.metric == "cosine":
                self.clamp_low = float("-inf")
            else:
                raise ValueError(f"Unknown metric {self.metric}")
        else:
            self.clamp_low = clamp_low
        if clamp_high == "auto":
            if self.metric == "euclidean":
                self.clamp_high = 1.0
            elif self.metric == "cosine":
                self.clamp_high = float("inf")
            else:
                raise ValueError(f"Unknown metric {self.metric}")
        else:
            self.clamp_high = clamp_high

        self.s = s
        self.Z_bar = Z_bar
        self.spec_param = noise_in_estimator
        self.neg_spec = neg_spec
        self.ince_spec = ince_spec

        self.early_exaggeration = early_exaggeration

        if self.early_exaggeration:
            if (self.loss_mode not in  ["infonce", "neg"]):
                print("Warning: Early exaggeration is only supported for loss modes 'infonce' and 'neg'.")

        # move to correct device at init, esp before registering with the optimizer
        self.model = self.model.to(self.device)

    def process_spec_param(self, n=None, X=None, s=None, Z_bar=None, neg_spec=None, ince_spec=None, overwrite=False):
        s = self.s if s is None else s
        Z_bar = self.Z_bar if Z_bar is None else Z_bar
        neg_spec = self.neg_spec if neg_spec is None else neg_spec
        ince_spec = self.ince_spec if ince_spec is None else ince_spec

        spec_param = 0.0  # just to set default in case it does not matter, e.g., for nce

        assert n is not None or X is not None, "Either n or X must be passed to process_spec_param"
        if self.loss_mode == "neg":
            # if not explicitly passed, use dataset length, only works if DataLoader is over data points, not over similar pairs
            n = len(X) if n is None else n
            if s is not None:
                # overwrite self.Z_bar
                Z_umap = n ** 2 / self.negative_samples
                Z_tsne = 100 * n
                # s = 0 --> z_tsne, s = 1 --> z_umap; using logs for numerical stability
                Z_bar = Z_tsne * np.exp(s * (np.log(Z_umap) - np.log(Z_tsne)))

            if Z_bar is not None:
                # assume uniform noise distribution over n**2 many edges
                neg_spec = self.negative_samples * Z_bar / n ** 2

            if neg_spec is not None:
                spec_param = neg_spec

        elif self.loss_mode == "infonce" or self.loss_mode == "infonce_alt":
            if s is not None:
                # overwrite self.spec_param, which will be passed to the loss
                # s=0 --> tsne (self.ince_param=1), s=1 --> umap-like (self.ince_param=4); logarithmic interpolation
                ince_spec = np.exp(s * np.log(4.0))

            # for passing to the loss
            if ince_spec is not None:
                spec_param = ince_spec

        if overwrite:
            self.spec_param = spec_param
            self.s = s
            self.Z_bar = Z_bar
            self.neg_spec = neg_spec
            self.ince_spec = ince_spec

        return spec_param

    def setup_optimizer(self):
        """
        Set up the optimizer
        :return
        """
        params = [{"params": self.model.parameters()}]
        if self.loss_mode == "nce":
            params += [
                {"params": self.log_Z, "lr": 0.001}
            ]  # make sure log_Z always has a sufficiently small lr

        if self.optimizer == "sgd":
            optimizer = torch.optim.SGD(
                params,
                lr=self.learning_rate,
                momentum=self.momentum,
                weight_decay=self.weight_decay,
            )
        elif self.optimizer == "adam":
            optimizer = torch.optim.Adam(
                params,
                weight_decay=self.weight_decay,
                lr=self.learning_rate,
            )
        else:
            raise ValueError(
                f"Only optimizer 'adam' and 'sgd' allowed, but is {self.optimizer}."
            )
        return optimizer



    def fit(self, X: torch.utils.data.DataLoader, n: int = None):
        """
        Train the model
        :param X: torch.utils.data.DataLoader Loads pairs of similar objects
        :param n: int Size of the dataset
        :return: self
        """
        # translate various spectrum parameters to self.spec_param.
        self.process_spec_param(n=n, X=X, overwrite=True)

        # set up loss
        criterion = ContrastiveLoss(
            negative_samples=self.negative_samples,
            metric=self.metric,
            temperature=self.temperature,
            loss_mode=self.loss_mode,
            spec_param=torch.tensor(self.spec_param).to(self.device),
            eps=torch.tensor(self.eps).to(self.device),
            clamp_high=self.clamp_high,
            clamp_low=self.clamp_low,
            seed=self.seed,
            loss_aggregation=self.loss_aggregation,
        )

        # set up optimizer
        optimizer = self.setup_optimizer()

        # initial callback
        if (
            self.save_freq is not None
            and self.save_freq > 0
            and callable(self.callback)
        ):
            self.callback(
                -1,
                self.model,
                self.negative_samples,
                self.loss_mode,
                self.log_Z,
            )

        batch_losses = []

        # logging memory usage
        mem_dict = {
            "active_bytes.all.peak": [],
            "allocated_bytes.all.peak": [],
            "reserved_bytes.all.peak": [],
            "reserved_bytes.all.allocated": [],
        }
        if self.early_exaggeration:
            s_early = max(self.s, 1.0) if self.s is not None else 1.0
            spec_param_early = self.process_spec_param(n=n, X=X, s=s_early)


        # training
        for epoch in range(self.n_epochs):
            if "cuda" in self.device:
                info = torch.cuda.memory_stats(self.device)
                [mem_dict[k].append(info[k]) for k in mem_dict.keys()]

            # select the correctly exaggerated spectrum parameter for this epoch
            if self.early_exaggeration and self.s is not None and epoch < self.n_epochs // 3:
                cur_spec_param = spec_param_early
            else:
                cur_spec_param = self.spec_param

            # update the spec param in the loss
            criterion.spec_param = torch.tensor(cur_spec_param).to(self.device)

            # first time after early exaggeration, reset the optimizer if it is adam
            #if self.early_exaggeration and epoch == self.n_epochs // 3 and self.optimizer == "adam":
            #    optimizer = self.setup_optimizer()

            # anneal learning rate
            lr = new_lr(
                self.learning_rate,
                self.anneal_lr,
                self.lr_decay_rate,
                lr_min_factor=self.lr_min_factor,
                cur_epoch=epoch,
                total_epochs=self.n_epochs,
                decay_epochs=self.lr_decay_epochs,
                warmup_epochs=self.warmup_epochs,
                warmup_lr=self.warmup_lr,
            )

            # just change the lr of the first param group, not that of Z
            optimizer.param_groups[0]["lr"] = lr

            # train for one epoch
            bl = train(
                X,
                self.model,
                self.log_Z,
                criterion,
                optimizer,
                epoch,
                clip_grad=self.clip_grad,
                print_freq=self.print_freq_iteration,
                force_resample=self.force_resample,
            )
            batch_losses.append(bl)

            # callback
            if (
                self.save_freq is not None
                and self.save_freq > 0
                and epoch % self.save_freq == 0
                and callable(self.callback)
            ):
                self.callback(
                    epoch, self.model, self.negative_samples, self.loss_mode, self.log_Z
                )
            # print epoch progress
            if self.print_freq_epoch is not None and epoch % self.print_freq_epoch == 0:
                print(f"Finished epoch {epoch}/{self.n_epochs}, loss {sum(bl)/ len(bl):.3f}", file=sys.stderr)

        self.losses = batch_losses
        self.mem_dict = mem_dict
        self.embedding_ = None
        return self

    def fit_transform(self, X):
        self.fit(X)
        return self.embedding_


class ContrastiveLoss(torch.nn.Module):
    """Supervised Contrastive Learning: https://arxiv.org/pdf/2004.11362.pdf.
    It also supports the unsupervised contrastive loss in SimCLR"""

    def __init__(
        self,
        negative_samples=5,
        temperature=0.07,
        loss_mode="all",
        metric="euclidean",
        base_temperature=1,
        eps=1.0,
        spec_param=1.0,
        clamp_high=1.0,
        clamp_low=1e-4,
        seed=0,
        loss_aggregation="mean",
    ):
        super(ContrastiveLoss, self).__init__()
        self.negative_samples = negative_samples
        self.temperature = temperature
        self.loss_mode = loss_mode
        self.metric = metric
        self.base_temperature = base_temperature
        self.spec_param = spec_param
        self.eps = eps
        self.clamp_high = clamp_high
        self.clamp_low = clamp_low
        self.seed = seed
        torch.manual_seed(self.seed)
        self.neigh_inds = None
        self.loss_aggregation = loss_aggregation

    def forward(self, features, log_Z=None, force_resample=False):
        """Compute loss for model. SimCLR unsupervised loss:
        https://arxiv.org/pdf/2002.05709.pdf

        Args:
            features: hidden vector of shape [2 * bsz, n_views, ...].
            log_Z: scalar, logarithm of the learnt normalization constant for nce.
            force_resample: Whether the negative samples should be forcefully resampled.
        Returns:
            A loss scalar.
        """

        batch_size = features.shape[0] // 2
        b = batch_size

        # We can at most sample this many samples from the batch.
        # `b` can be lower than `self.negative_samples` in the last batch.
        negative_samples = min(self.negative_samples, 2 * b - 1)

        if force_resample or self.neigh_inds is None:
            neigh_inds = make_neighbor_indices(
                batch_size, negative_samples, device=features.device
            )
            self.neigh_inds = neigh_inds
        # # untested logic to accomodate for last batch
        # elif self.neigh_inds.shape[0] != batch_size:
        #     neigh_inds = make_neighbor_indices(batch_size, negative_samples)
        #     # don't save this one
        else:
            neigh_inds = self.neigh_inds
        neighbors = features[neigh_inds]

        # `neigh_mask` indicates which samples feel attractive force
        # and which ones repel each other
        neigh_mask = torch.ones_like(neigh_inds, dtype=torch.bool)
        neigh_mask[:, 0] = False

        origs = features[:b]

        # compute probits
        if self.metric == "euclidean":
            dists = ((origs[:, None] - neighbors) ** 2).sum(axis=2)
            # Cauchy affinities
            probits = torch.div(1, self.eps + dists)
        elif self.metric == "cosine":
            norm = torch.nn.functional.normalize
            o = norm(origs.unsqueeze(1), dim=2)
            n = norm(neighbors.transpose(1, 2), dim=1)
            logits = torch.bmm(o, n).squeeze() / self.temperature
            # logits_max, _ = logits.max(dim=1, keepdim=True)
            # logits -= logits_max.detach()
            # logits -= logits.max().detach()
            probits = torch.exp(logits)
        else:
            raise ValueError(f"Unknown metric “{self.metric}”")

        # compute loss
        if self.loss_mode == "nce":
            # for proper nce it should be negative_samples * p_noise. But for
            # uniform noise distribution we would need the size of the dataset
            # here. Also, we do not use a uniform noise distribution as we sample
            # negative samples from the batch.

            if self.metric == "euclidean":
                # estimator is (cauchy / Z) / ( cauchy / Z + neg samples)). For numerical
                # stability rewrite to 1 / ( 1 + (d**2 + eps) * Z * m)
                estimator = 1 / (
                    1 + (dists + self.eps) * torch.exp(log_Z) * negative_samples
                )
            else:
                probits = probits / torch.exp(log_Z)
                estimator = probits / (probits + negative_samples)

            loss = -(~neigh_mask * torch.log(estimator.clamp(self.clamp_low, self.clamp_high))) - (
                neigh_mask * torch.log((1 - estimator).clamp(self.clamp_low, self.clamp_high))
            )
        elif self.loss_mode == "neg":
            if self.metric == "euclidean":
                # estimator rewritten for numerical stability as for nce
                estimator = 1 / (1 + self.spec_param * (dists + self.eps))
            else:
                estimator = probits / (probits + self.spec_param)

            loss = -(~neigh_mask * torch.log(estimator.clamp(self.clamp_low, self.clamp_high))) - (
                neigh_mask * torch.log((1 - estimator).clamp(self.clamp_low, self.clamp_high))
            )

        elif self.loss_mode == "umap":
            # cross entropy parametric umap loss
            loss = -(~neigh_mask * torch.log(probits.clamp(self.clamp_low, self.clamp_high))) - (
                neigh_mask * torch.log((1 - probits).clamp(self.clamp_low, self.clamp_high))
            )
        elif self.loss_mode == "infonce":
            # loss from e.g. sohn et al 2016, includes pos similarity in denominator
            loss = -(self.temperature / self.base_temperature) * (
                (torch.log(probits.clamp(self.clamp_low, self.clamp_high)[~neigh_mask]))
                - self.spec_param**(-1) * torch.log(probits.clamp(self.clamp_low, self.clamp_high).sum(axis=1))
            )
        elif self.loss_mode == "infonce_alt":
            # loss simclr
            loss = -(self.temperature / self.base_temperature) * (
                (torch.log(probits.clamp(self.clamp_low, self.clamp_high)[~neigh_mask]))
                - self.spec_param**(-1) * torch.log((neigh_mask *
                                                       probits.clamp(self.clamp_low, self.clamp_high)).sum(axis=1))
            )
        else:
            raise ValueError(f"Unknown loss_mode “{self.loss_mode}”")

        # aggregate loss over batch
        if self.loss_aggregation == "sum":
            loss = loss.sum()
        else:
            loss = loss.mean()

        return loss


def new_lr(
    learning_rate,
    anneal_lr,
    lr_decay_rate,
    lr_min_factor,
    cur_epoch,
    total_epochs,
    decay_epochs=None,  # unused for now
    warmup_lr=0,
    warmup_epochs=0,
):
    """
    Decays the learning rate
    :param learning_rate: float Current learning rate
    :param anneal_lr: str Specifies the learning rate annealing. Must be one of "none", "linear" or "cosine"
    :param lr_decay_rate: float Rate of cosine decay.
    :param lr_min_factor: float Minimal learning rate of linear decay.
    :param cur_epoch: int Current epoch
    :param total_epochs: int Total number of epochs
    :param decay_epochs: int Number of decay epochs (unused)
    :param warmup_epochs: int Number of epochs for linearly warming up the learning rate
    :param warmup_lr: float Starting learning rate to warm up from.
    :return: float New learning rate
    """
    anneal_epochs = total_epochs - warmup_epochs
    if cur_epoch < warmup_epochs:
        lr = warmup_lr + (learning_rate - warmup_lr) * cur_epoch / warmup_epochs
    else:
        cur_epoch = cur_epoch - warmup_epochs
        if anneal_lr == "none":
            lr = learning_rate
        elif anneal_lr == "linear":
            lr = learning_rate * max(lr_min_factor, 1 - cur_epoch / anneal_epochs)
        elif anneal_lr == "cosine":
            eta_min = 0
            lr = (
                eta_min
                + (learning_rate - eta_min)
                * (1 + np.cos(np.pi * cur_epoch / anneal_epochs))
                / 2
            )
        else:
            raise RuntimeError(f"Unknown learning rate annealing “{anneal_lr = }”")

    return lr


def make_neighbor_indices(batch_size, negative_samples, device=None):
    """
    Selects neighbor indices
    :param batch_size: int Batch size
    :param negative_samples: int Number of negative samples
    :param device: torch.device Device of the model
    :return: torch.tensor Neighbor indices
    :rtype:
    """
    b = batch_size

    if negative_samples < 2 * b - 2:
        # uniform probability for all points in the minibatch,
        # we sample points for repulsion randomly
        neg_inds = torch.randint(0, 2 * b - 1, (b, negative_samples), device=device)
        neg_inds += (torch.arange(1, b + 1, device=device) - 2 * b)[:, None]
    else:
        # full batch repulsion
        all_inds1 = torch.repeat_interleave(
            torch.arange(b, device=device)[None, :], b, dim=0
        )
        not_self = ~torch.eye(b, dtype=bool, device=device)
        neg_inds1 = all_inds1[not_self].reshape(b, b - 1)

        all_inds2 = torch.repeat_interleave(
            torch.arange(b, 2 * b, device=device)[None, :], b, dim=0
        )
        neg_inds2 = all_inds2[not_self].reshape(b, b - 1)
        neg_inds = torch.hstack((neg_inds1, neg_inds2))

    # now add transformed explicitly
    neigh_inds = torch.hstack(
        (torch.arange(b, 2 * b, device=device)[:, None], neg_inds)
    )

    return neigh_inds





import cne
import numpy as np

# get random data
x1 = np.random.randn(1000, 10)
x2 = np.random.randn(1000, 10) + np.ones((1000, 10))
x = np.concatenate([x1, x2])

# non-parametric Neg-t-SNE
#embedder_neg = cne.CNE(loss_mode="nce",
#                       k=15,
#                       optimizer="sgd",
#                       parametric=False,
#                       #print_freq_in_epoch=None,
#                       #learning_rate=0.1,
#                       print_freq_epoch=10)
# embd_neg = embedder_neg.fit_transform(x)
# print("Test OK!")


from cne import CNE
C_128 = CNE(parametric=False, embd_dim=2, loss_mode="neg")
Z = C_128.fit_transform(x)


import torch
import numpy as np

from .cne import ContrastiveEmbedding
from annoy import AnnoyIndex
from scipy.sparse import lil_matrix
from sklearn.decomposition import PCA
import time

# various datasets / dataloaders
class NeighborTransformData(torch.utils.data.Dataset):
    """Returns a pair of neighboring points in the dataset."""
    def __init__(
            self, dataset, neighbor_mat, random_state=None
    ):
        self.dataset = torch.tensor(dataset, dtype=torch.float32)
        self.neighbor_mat = neighbor_mat
        self.rng = np.random.default_rng(random_state)

    def __len__(self):
        return len(self.dataset)

    def __getitem__(self, i):
        neighs = self.neighbor_mat[i].nonzero()[1]
        nidx = self.rng.choice(neighs)

        item = self.dataset[i]
        neigh = self.dataset[nidx]
        return item, neigh


class NeighborTransformIndices(torch.utils.data.Dataset):
    """Returns a pair of indices of neighboring points in the dataset."""
    def __init__(
            self, neighbor_mat, random_state=None
    ):
        neighbor_mat = neighbor_mat.tocoo()
        self.heads = torch.tensor(neighbor_mat.row)
        self.tails = torch.tensor(neighbor_mat.col)

    def __len__(self):
        return len(self.heads)

    def __getitem__(self, i):
        return self.heads[i], self.tails[i]



class NumpyToTensorDataset(torch.utils.data.Dataset):
    def __init__(self, dataset, reshape=None):
        self.dataset = torch.tensor(dataset, dtype=torch.float32)
        if reshape is not None:
            self.reshape = lambda x: np.reshape(x, reshape)
        else:
            self.reshape = lambda x: x

    def __len__(self):
        return len(self.dataset)

    def __getitem__(self, i):
        item = self.dataset[i]
        return self.reshape(item)


class NumpyToIndicesDataset(torch.utils.data.Dataset):
    def __init__(self, n):
        self.n = n

    def __len__(self):
        return self.n

    def __getitem__(self, i):
        return i


# based on https://discuss.pytorch.org/t/dataloader-much-slower-than-manual-batching/27014/6
class FastTensorDataLoader:
    """
    A DataLoader-like object for a set of tensors that can be much faster than
    TensorDataset + DataLoader because dataloader grabs individual indices of
    the dataset and calls cat (slow).
    """
    def __init__(self, neighbor_mat, batch_size=1024, shuffle=False, data_on_gpu=False, drop_last=False, seed=0):
        """
        Initialize a FastTensorDataLoader.

        :param *tensors: tensors to store. Must have the same length @ dim 0.
        :param batch_size: batch size to load.
        :param shuffle: if True, shuffle the data *in-place* whenever an
            iterator is created out of this object.
        :param data_on_gpu: If True, the dataset is loaded on GPU as a whole.
        :param drop_last: Drop the last batch if it is smaller than the others.
        :param seed: Random seed

        :returns: A FastTensorDataLoader.
        """

        neighbor_mat = neighbor_mat.tocoo()
        tensors = [torch.tensor(neighbor_mat.row),
                   torch.tensor(neighbor_mat.col)]
        assert all(t.shape[0] == tensors[0].shape[0] for t in tensors)

        # manage device
        if data_on_gpu:
            self.device = "cuda"
            tensors = [tensor.to(self.device) for tensor in tensors]
        else:
            self.device = "cpu"
        self.tensors = tensors

        self.dataset_len = self.tensors[0].shape[0]
        self.batch_size = batch_size
        self.shuffle = shuffle
        self.drop_last = drop_last
        self.seed = seed
        torch.manual_seed(self.seed)

        # Calculate number of  batches
        n_batches, remainder = divmod(self.dataset_len, self.batch_size)
        if remainder > 0 and not self.drop_last:
            n_batches += 1
        self.n_batches = n_batches

        self.batch_size = torch.tensor(self.batch_size, dtype=int).to(self.device)

    def __iter__(self):
        if self.shuffle:
            self.indices = torch.randperm(self.dataset_len, device=self.device)
        else:
            self.indices = None
        self.i = 0
        return self

    def __next__(self):
        if self.i > self.dataset_len - self.batch_size:
            raise StopIteration
        if self.indices is not None:
            indices = self.indices[self.i:self.i+self.batch_size]
            batch = tuple(torch.index_select(t, 0, indices) for t in self.tensors)
        else:
            batch = tuple(t[self.i:self.i+self.batch_size] for t in self.tensors)
        self.i += self.batch_size
        return batch

    def __len__(self):
        return self.n_batches


class FCNetwork(torch.nn.Module):
    "Fully-connected network"
    def __init__(self, in_dim=784, feat_dim=2):
        super(FCNetwork, self).__init__()
        self.flatten = torch.nn.Flatten()
        self.linear_relu_stack = torch.nn.Sequential(
            torch.nn.Linear(in_dim, 100),
            torch.nn.ReLU(),
            torch.nn.Linear(100, 100),
            torch.nn.ReLU(),
            torch.nn.Linear(100, 100),
            torch.nn.ReLU(),
            torch.nn.Linear(100, feat_dim),
        )

    def forward(self, x):
        x = self.flatten(x)
        logits = self.linear_relu_stack(x)
        return logits


class CNE(object):
    """
    Manages contrastive neighbor embeddings.
    """
    def __init__(self,
                 model=None,
                 k=15,
                 parametric=False,
                 data_on_gpu="auto",
                 seed=0,
                 anneal_lr=True,
                 embd_dim=2,
                 **kwargs):
        """
        :param model: Embedding model
        :param k: int Number of nearest neighbors
        :param parametric: bool If True and model=None uses a parametric embedding model
        :param data_on_gpu: bool or "auto" Load whole dataset to GPU and try to use pykeops for kNN graph if possible.
        :param seed: int Random seed
        :param anneal_lr: bool If True anneal the learning rate linearly.
        :param kwargs:
        """
        self.model = model
        self.k = k
        self.parametric = parametric
        self.data_on_gpu = data_on_gpu
        self.kwargs = kwargs
        self.seed = seed
        self.anneal_lr = anneal_lr
        self.embd_dim = embd_dim


    def fit_transform(self, X, init=None, graph=None):
        "Fit the model, then transform."
        self.fit(X, init=init, graph=graph)
        return self.transform(X, fit_transform=True)

    def transform(self, X, fit_transform=False):
        "Transform a dataset using the fitted model."
        if self.parametric:
            X = X.reshape(X.shape[0], -1)
            self.dataset_plain = NumpyToTensorDataset(X)
            self.dl_unshuf = torch.utils.data.DataLoader(
                self.dataset_plain,
                shuffle=False,
                batch_size=self.cne.batch_size,
            )
            model = self.network
            device = self.cne.device
            embd = np.vstack([model(batch.to(device))
                            .detach().cpu().numpy()
                            for batch in self.dl_unshuf])
        else:
            embd = self.model.weight.detach().cpu().numpy()
            if isinstance(X, int):
                embd = embd[X]
            elif isinstance(X, np.ndarray) and len(np.squeeze(X).shape) == 1:
                if X.dtype == int:
                    embd = embd[np.squeeze(X)]
            elif isinstance(X, list) and np.all([isinstance(x, int) for x in X]):
                embd = embd[X]
            else:
                if not fit_transform:
                    print("Warning: A non-parametric model cannot transform new data. Returning the embedding of the training data. "
                          "Pass an integer (or a list / np.array thereof) to obtain the corresponding training embeddings")
        return embd

    def fit(self, X, init=None, graph=None):
        """
        Fit the model
        :param X: np.array Dataset
        :param init: np.array Initial embedding. If None, use PCA rescaled so that first PC has standard deviation 1.
        :param graph: graph encoding similarity. If None, a kNN graph will be computed. This is done with pykeops if the
        ContrastiveEmbedding instance is on GPU, otherwise annoy is used. Passing "annoy" or "pykeops" forces to use this library for
         the kNN graph computation. Pykeops requires a GPU but is much faster. Alternatively, any scipy.sparse.csr_matrix
         can be passed.
        :return:
        """
        start_time = time.time()
        X = X.reshape(X.shape[0], -1)
        in_dim = X.shape[1]
        # set up model if not given
        if self.model is None:
            if self.parametric:
                self.embd_layer = torch.nn.Embedding.from_pretrained(torch.tensor(X).to(torch.float32),
                                                                     freeze=True)
                self.network = FCNetwork(in_dim, feat_dim=self.embd_dim)
                self.model = torch.nn.Sequential(
                    self.embd_layer,
                    self.network
                )
            else:
                if init is None:
                    # default to pca
                    pca_projector = PCA(n_components=self.embd_dim)
                    init = pca_projector.fit_transform(X)
                    init /= (init[:, 0].std())
                elif type(init).__module__ == np.__name__:
                    assert len(init) == len(X), f"Data and initialization must have the same number of elements but have {len(X)} and {len(init)}."
                    assert len(init.shape) == 2, f"Initialization must have 2 dimensions but has {len(init.shape)}."
                    if init.shape[1] != self.embd_dim:
                        print(f"Warning: Initialization has {init.shape[1]} dimensions but {self.embd_dim} are expected."
                              f" Will use the initialization's {init.shape[1]} dimensions.")
                # All embedding parameters will be part of the model. This is
                # conceptually easy, but limits us to embeddings that fit on the
                # GPU.
                self.model = torch.nn.Embedding.from_pretrained(torch.tensor(init))
                self.model.requires_grad_(True)

        # use higher learning rate for non-parametric version
        if "learning_rate" not in self.kwargs.keys():
            lr = 0.001 if self.parametric else 1.0
            self.kwargs["learning_rate"] = lr

        # Load embedding engine
        self.cne = ContrastiveEmbedding(self.model,
                                        seed=self.seed,
                                        anneal_lr=self.anneal_lr,
                                        **self.kwargs)

        # is no graph is passed, compute the similarity graph with pykeops is cuda is available and otherwise annoy
        if graph is None:
            # select annoy or pykeops depending on data_on_gpu and availability of pykeops
            if self.cne.device == "cuda":
                try:
                    import pykeops
                    graph = "pykeops"
                except ImportError:
                    graph = "annoy"
            else:
                graph = "annoy"

        if isinstance(graph, str):
            if graph == "annoy":
                print("Computing approximate kNN graph with annoy")
                # create approximate NN search tree
                self.annoy = AnnoyIndex(in_dim, "euclidean")
                [self.annoy.add_item(i, x) for i, x in enumerate(X)]
                self.annoy.build(50)

                # construct the adjacency matrix for the graph
                adj = lil_matrix((X.shape[0], X.shape[0]))
                for i in range(X.shape[0]):
                    neighs_, _ = self.annoy.get_nns_by_item(i, self.k + 1, include_distances=True)
                    neighs = neighs_[1:]
                    adj[i, neighs] = 1
                    adj[neighs, i] = 1  # symmetrize on the fly

                self.neighbor_mat = adj.tocsr()
            elif graph == "pykeops":
                print("Computing approximate kNN graph with pykeops")
                from pykeops.torch import LazyTensor
                import scipy.sparse

                # set up pykeops LazyTensors
                x_cuda = torch.tensor(X).to("cuda").contiguous()
                x_i = LazyTensor(x_cuda[:, None])
                x_j = LazyTensor(x_cuda[None])

                # compute distance and knn_idx with keops
                dists = ((x_i - x_j) ** 2).sum(-1)
                knn_idx = dists.argKmin(K=self.k + 1, dim=0)[:, 1:].cpu().numpy().flatten()

                # construct the adjacency matrix for the graph
                knn_graph = scipy.sparse.coo_matrix((np.ones(len(X) * self.k),
                                                     (np.repeat(np.arange(X.shape[0]), self.k),
                                                      knn_idx)),
                                                    shape=(len(X), len(X)))

                # symmetrize on the fly
                self.neighbor_mat = knn_graph.maximum(knn_graph.transpose()).tocsr()
            else:
                raise ValueError("When passing a string as graph it must be 'annoy' or 'pykeops'")
        else:
            self.neighbor_mat = graph.tocsr()

        if self.data_on_gpu == "auto":
            self.data_on_gpu = self.cne.device == "cuda"
        if self.data_on_gpu and self.cne.device == "cpu":
            print("Warning: Data is on GPU but the model is on CPU. This will be unnecessarily slow.")

        # create data loader
        self.dataloader = FastTensorDataLoader(self.neighbor_mat,
                                               shuffle=True,
                                               batch_size=self.cne.batch_size,
                                               data_on_gpu=self.data_on_gpu,
                                               seed=self.seed)

        # fit the model
        self.cne.fit(self.dataloader, len(X))
        end_time = time.time()
        self.cne.time = end_time - start_time
        return self



import torch
import numpy as np
try:
    from vis_utils.utils import compute_normalization, expected_loss_keops, \
        NCE_loss_keops, KL_divergence
    vis_utils_available = True
except:
    vis_utils_available = False

class Logger():
    """
    Class for logging various quantities of interest during a contrastive neighbor embedding optimization.
    """
    def __init__(self,
                 log_embds=False,
                 log_losses=False,
                 graph=None,
                 log_norms=False,
                 log_kl=False,
                 n=None):
        """
        :param log_embds: bool If true, log intermediate embeddings.
        :param log_losses: bool If true, log the (expected) loss of the model.
        :param graph: sparse matrix Holds the similarity graph.
        :param log_norms: bool If true, log the intermediate values of the parition function.
        :param log_kl: bool If true, log the intermediate values of the KL divergence.
        :param n: int Dataset size
        """

        self.log_embds = log_embds
        if self.log_embds:
            self.embds = []
            self.Zs = []
        else:
            self.embds = None

        self.graph = graph.tocoo() if graph is not None else graph

        self.log_losses = log_losses
        if self.log_losses:
            assert vis_utils_available, f"Need vis_utils package to log losses."
            self.losses = []

        self.log_kl = log_kl
        if self.log_kl:
            assert vis_utils_available
            assert self.graph is not None
            self.kls = []

        self.log_norm = log_norms
        if self.log_norm:
            assert vis_utils_available, "Need vis_utils package to log normalization constant."
            self.norms = []

        if n is not None:
            self.ds = torch.utils.data.TensorDataset(torch.arange(n))
            self.dl = torch.utils.data.DataLoader(self.ds,
                                                  batch_size=256,
                                                  shuffle=False)

    def __call__(self, epoch, model, negative_samples, loss_mode, log_Z=None, noise_in_estimator=None):
        # read out the embeddings from the model if anything shall be logged
        if isinstance(model, torch.nn.modules.sparse.Embedding):
            # non-parametric case, just get all embeddings from embedding layer
            embd = model.weight.detach().cpu().numpy()
        else:
            # parametric case, model is Embedding layer + FCNetwork
            # Just feed indices from self.dl
            device = model[0].weight.device
            embd = np.vstack([model(batch[0].to(device))
                             .detach().cpu().numpy()
                              for batch in self.dl])
        if log_Z is not None and self.log_embds:
            self.Zs.append(torch.exp(log_Z).detach().cpu().numpy())

        if self.log_embds:
            self.embds.append(embd)
        else:
            self.embds = [embd]

        if self.log_losses:
            if loss_mode == "umap":
                self.losses.append(expected_loss_keops(high_sim=self.graph,
                                                       embedding=embd,
                                                       a=1.0,
                                                       b=1.0,
                                                       negative_sample_rate=negative_samples,
                                                       push_tail=True))
            elif loss_mode == "nce":
                # transform Z since the implemented criterion does not consider the noise distribution explicitly
                Z_prime = np.exp(log_Z.detach().cpu().numpy())
                Z_prime = Z_prime * len(embd)**2
                self.losses.append(NCE_loss_keops(high_sim=self.graph,
                                                  embedding=embd,
                                                  m=negative_samples,
                                                  Z=Z_prime,
                                                  a=1.0,
                                                  b=1.0,
                                                  noise_log_arg=True,
                                                  eps=1e-4))
            elif loss_mode == "neg":
                # turn noise_in_estimator back into Z via
                # Z * m * p_n = 1 <--> Z = 1 / (m * p_n)
                Z = (negative_samples / len(embd)**2)**-1
                self.losses.append(NCE_loss_keops(high_sim=self.graph,
                                                  embedding=embd,
                                                  m=negative_samples,
                                                  Z=Z,
                                                  a=1.0,
                                                  b=1.0,
                                                  noise_log_arg=True,
                                                  eps=1e-4))
            else:
                raise NotImplementedError("No expected loss function is implemented for InfoNCE")

        if self.log_kl:
            self.kls.append(KL_divergence(self.graph,
                                          embedding=embd,
                                          a=1.0,
                                          b=1.0,
                                          sim_func="cauchy",
                                          eps = float(np.finfo(float).eps),
                                          norm_over_pos=False)
                            )
        if self.log_norm:
            norm = compute_normalization(embd,
                                         sim_func="cauchy",
                                         no_diag=True,
                                         a=1.0,
                                         b=1.0,
                                         eps=float(np.finfo(float).eps))
            self.norms.append(norm.detach().cpu().numpy())



